import typing

from ._thread_env import _THREAD_ENV
from .base import CurlStreamHandlerBase


if _THREAD_ENV == "gevent":
	from .gevent_handler import CurlStreamHandlerGevent
	_Impl = CurlStreamHandlerGevent
else:
	# Alternatively, CurlStreamHandlerThreads
	from .multi_handler import CurlStreamHandlerMulti
	_Impl = CurlStreamHandlerMulti

CurlStreamHandler: typing.Type[CurlStreamHandlerBase] = _Impl


import typing
from typing import TYPE_CHECKING
import warnings

import pycurl
import gevent
from gevent.event import AsyncResult
from gevent.lock import Semaphore

if TYPE_CHECKING:
	from gevent._types import _IoWatcher # type: ignore


GEVENT_READ = 1
GEVENT_WRITE = 2

class GeventPyCurl:
	'''
		Usage:

		multi_curl = GeventPyCurl()
		result = multi_curl.add_handle(curl_handle)

		result.wait()
	'''

	def __init__(self):
		"""
		Parameters:
			cacert: CA cert path to use, by default, certs from ``certifi`` are used.
			loop: EventLoop to use.
		"""
		self._curl_multi = pycurl.CurlMulti()
		
		self.loop = gevent.get_hub().loop

		self._timer: typing.Optional[gevent.Greenlet] = None
		self._watchers: dict[typing.Any, dict[str, _IoWatcher]] = {}

		self._results: dict[pycurl.Curl, AsyncResult] = {}
		self._callbacks: dict[pycurl.Curl, callable] = {}

		self._checker: gevent.Greenlet = None #self._checker = gevent.spawn(self._force_timeout)
		
		self._start_closing = False

		self._set_options()
	
	def add_handle(self, curl: pycurl.Curl, cleanup_after_perform: typing.Callable[[typing.Optional[Exception]], None]=None):
		"""Add a curl handle to be managed by curl_multi. This is the equivalent of
		`perform` in the async world."""

		if self._start_closing:
			raise RuntimeError("This curl_multi instance is closed.")

		self._curl_multi.add_handle(curl)
	
		result = AsyncResult()
		self._results[curl] = result
		self._callbacks[curl] = cleanup_after_perform

		return result
	
	def cancel_handle(self, curl: pycurl.Curl):
		"""Cancel is not natively supported in gevent.AsyncResult."""
		self._set_exception(curl, RuntimeError("Cancelled"))
	
	def graceful_close(self):
		self._start_closing = True

	def close(self):
		"""Close and cleanup running timers, readers, writers and handles."""
		 # Close and wait for the force timeout checker to complete

		if self._checker and not self._checker.dead:
			self._checker.kill(block=False)

		# Close all pending futures (if any)
		for curl in list(self._results.keys()):
			self.cancel_handle(curl)
			
		# Cleanup curl_multi handle
		self._curl_multi.close()
		self._curl_multi = None

		# Remove watchers
		for sockfd, entry in list(self._watchers.items()):
			if entry.get("watcher"):
				entry.get("watcher").stop()      # stop monitoring
				entry.get("watcher").close()      # dispose of the watcher
				del self._watchers[sockfd]

		# Kill timer
		if self._timer is not None:
			self._timer.kill(block=False)
			self._timer = None
	
	def _timer_function(self, timeout_ms: int):
		
		if self._timer:
			self._timer.kill(block=False)
			self._timer = None

		if not self._curl_multi:
			# Don't schedule timers if multi handle is gone.
			return 0

		if timeout_ms == -1:
			# A timeout_ms value of -1 means you should delete the timer, we already did.
			return 0

		if timeout_ms == 0:
			self._timer = gevent.spawn(
				self._process_data, 
				pycurl.SOCKET_TIMEOUT, 
				pycurl.POLL_NONE
			)
		elif timeout_ms > 0:
				# spawn a greenlet to run after timeout_ms milliseconds
				self._timer = gevent.spawn_later(
					timeout_ms / 1000.0,
					self._process_data,
					pycurl.SOCKET_TIMEOUT,
					pycurl.POLL_NONE,
				)
		return 0
				
	def _socket_function(self, event, sockfd, multi, data):
		"""Called by libcurl to tell us what it wants on this fd."""

		want_read  = bool(event & pycurl.POLL_IN)
		want_write = bool(event & pycurl.POLL_OUT)

		# compute the new mask for gevent
		new_mask = 0
		if want_read:  new_mask |= GEVENT_READ
		if want_write: new_mask |= GEVENT_WRITE

		# teardown if libcurl says “remove”
		if event & pycurl.POLL_REMOVE:
			self._update_watcher(sockfd, 0)
			return

		# otherwise install/update the watcher
		self._update_watcher(sockfd, new_mask)

	def _update_watcher(self, fd: int, mask: int):
		"""
		Ensure there's exactly one I/O watcher for `fd` with the given mask.
		If mask==0, we stop+remove it. If mask changes, we stop+recreate.
		"""
		entry = self._watchers.get(fd)

		# nothing to do if mask didn’t change
		if entry and entry["mask"] == mask:
			return

		# stop old watcher if any
		if entry:
			entry["watcher"].stop()
			entry["watcher"].close()
			del self._watchers[fd]

		# if new mask is zero, we’re done
		if mask == 0:
			return

		# create a new watcher for read/write as needed
		w = self.loop.io(fd, mask, ref=True, priority=None)
		# callback only gets fd; we’ll re-derive the libcurl bitmask from mask
		w.start(self._on_watcher_event, fd)

		# stash both watcher _and_ the mask we asked for
		self._watchers[fd] = {"watcher": w, "mask": mask}

	def _on_watcher_event(self, fd: int):
		"""
		A gevent-watcher fired on `fd`.  Look up its mask and call curl.
		"""
		entry = self._watchers.get(fd)
		if not entry:
			return

		mask = entry["mask"]
		ev_bitmask = 0
		if mask & GEVENT_READ:
			ev_bitmask |= pycurl.CSELECT_IN
		if mask & GEVENT_WRITE:
			ev_bitmask |= pycurl.CSELECT_OUT

		# this is exactly what you were doing before:
		# call socket_action + info_read → _process_data
		self._process_data(fd, ev_bitmask)

	def _set_options(self):
		self._curl_multi.setopt(pycurl.M_TIMERFUNCTION, self._timer_function)
		self._curl_multi.setopt(pycurl.M_SOCKETFUNCTION, self._socket_function)

	def _socket_action(self, sockfd: int, ev_bitmask: int) -> int:
		"""Call libcurl _socket_action function"""
		ret, num_handles = self._curl_multi.socket_action(sockfd, ev_bitmask)
		return ret

	def _process_data(self, sockfd: int, ev_bitmask: int):
		"""Call curl_multi_info_read to read data for given socket."""
		if not self._curl_multi:
			warnings.warn(
				"Curlm already closed! quitting from _process_data",
				stacklevel=2,
			)
			return

		self._socket_action(sockfd, ev_bitmask)

		while True:
			if not self._curl_multi:
				break
			num_q, ok_list, err_list = self._curl_multi.info_read()
			for curl in ok_list:
				self._set_result(curl)

			for curl, errno, errmsg in err_list:
				curl_error = pycurl.error(errno, errmsg)
				self._set_exception(curl, curl_error)
			
			if num_q == 0:
				break
	
	def _force_timeout(self):
		while self._curl_multi:
			gevent.sleep(1)
			self._socket_action(pycurl.SOCKET_TIMEOUT, pycurl.POLL_NONE)
			
	def _callback(self, curl: pycurl.Curl, error: Exception=None):
		if curl in self._callbacks:
			callback = self._callbacks.pop(curl)
			if callable(callback):
				callback(error)
	
	def _pop_future(self, curl: pycurl.Curl):
		self._curl_multi.remove_handle(curl)
		return self._results.pop(curl, None)
	
	def _set_result(self, curl: pycurl.Curl):
		result = self._pop_future(curl)
		self._callback(curl)
		if result and not result.ready():
			result.set(None)

		if self._start_closing and not self._results:
			self.close()

	def _set_exception(self, curl: pycurl.Curl, exception):
		result = self._pop_future(curl)
		self._callback(curl, exception)
		if result and not result.ready():
			result.set_exception(exception)

		if self._start_closing and not self._results:
			self.close()


import traceback

import pycurl
import curl_cffi.curl

from curl_cffi.curl import CurlOpt
from curl_cffi._wrapper import ffi, lib

from .base import (
	CurlStreamHandlerBase, 
	QueueContinueRead, 
	QueueBreakRead
)

from ._thread_env import (
	_THREAD_ENV, _THREAD_CLASS, _THREAD_SLEEP, _THREAD_EVENT, _THREAD_QUEUE_MODULE
)


class CurlStreamHandlerMulti(CurlStreamHandlerBase):
	'''
		Curl Stream Handler (c) 2025 by Elis K.

		Using curl's multi interface and Python's coroutines.
	'''
	CURLMSG_DONE = 1

	def _perform_multi_curl_cffi(self):
		'''
			CurlCffi multi perform
		'''

		#  This functions reads & saves data to our queue, one chunk at a time. Non‐blocking pass over sockets.
		err_code = lib.curl_multi_perform(self.curl_multi, self.curl_multi_running_pointer)

		transfer_complete = False

		# Get any messages for curl handle
		if (not err_code or err_code == -1):
			
			# ask for the next timeout
			if self.curl_multi_running_pointer[0]:
				try:
					tptr = ffi.new("long *")
					lib.curl_multi_timeout(self.curl_multi, tptr)
					ms = tptr[0]
					if ms < 0:
						ms = 1000   # or whatever “default” poll interval you like
   					# block here until either a socket event *or* the timeout elapses
					nfds = ffi.new("int *", 0)
					lib.curl_multi_wait(self.curl_multi, ffi.NULL, 0, int(ms), nfds)
				except Exception as e:
					if self.debug:
						traceback.print_exc()

			# drain messages
			try:
				msgq = ffi.new("int *", 0)
				while True:
					msg = lib.curl_multi_info_read(self.curl_multi, msgq)
					if msg == ffi.NULL:
						break	
					# CURLMSG_DONE == transfer complete
					if msg.msg == self.CURLMSG_DONE:
						# translate the numeric code into a CurlError
						transfer_complete = True
						if msg.data.result != 0:
							self.error = self.curl._get_error(msg.data.result, "perform")
			except:
				if self.debug:
					traceback.print_exc()
		else:
			self.error = self.curl._get_error(err_code, "perform")
						
		if (
			transfer_complete
			or self.error
			or not self.curl_multi_running_pointer[0] 
			or self.quit_event.is_set()
		):

			if self.debug:
				print(f"[DEBUG] Closing curl. transfer_complete: {transfer_complete}, error: {bool(self.error)}, running: {bool(self.curl_multi_running_pointer[0])}, quit_event: {self.quit_event.is_set()}")
			
			self._cleanup_after_perform()
			
	def _perform_multi_pycurl(self):
		'''
			Pycurl multi perform
		'''
		err_code, running = self.curl_multi.perform()

		transfer_complete = False
		
		if running and (not err_code or err_code == pycurl.E_CALL_MULTI_PERFORM):
			
			# Blocking poll sleep before next call
			ms = self.curl_multi.timeout()
			if ms < 0:
				ms = 1000
			self.curl_multi.select(ms / 1000.0)

			# I don't know why we need to run perform again straight away after socket select, but it works for reading the error messages, so let's leave it like that
			err_code, running = self.curl_multi.perform()

		
		if (not err_code or err_code == pycurl.E_CALL_MULTI_PERFORM):
			try:
				# Check curl handle messages
				while True:
					num_q, ok_list, err_list = self.curl_multi.info_read()
					
					if len(ok_list):
						transfer_complete = True

					if len(err_list):
						# err_list is a list of (handle, errno, errmsg)
						_, errno, errmsg = err_list[0]
						self.error = pycurl.error(errno, errmsg)
					
					if num_q == 0:
						break
				
			except:
				if self.debug:
					traceback.print_exc()
		else:
			self.error = pycurl.error(err_code, "perform")
			

		if (
			transfer_complete
			or not running 
			or self.error
			or self.quit_event.is_set()
		):

			if self.debug:
				print(f"[DEBUG] Closing curl. transfer_complete: {transfer_complete}, error: {bool(self.error)}, running: {bool(running)}, quit_event: {self.quit_event.is_set()}")
			
			self._cleanup_after_perform()
	
	def _cleanup_after_perform(self):
		try:

			if self.curl_type == "pycurl":
				self.curl_multi.remove_handle(self.curl)
			
			elif self.curl_type == "curl_cffi":
				lib.curl_multi_remove_handle(self.curl_multi, self.curl._curl)

		except Exception:
			if self.debug:
				traceback.print_exc()
		finally:

			if self.curl_type == "pycurl":
				self.curl_multi.close()
			elif self.curl_type == "curl_cffi":
				lib.curl_multi_cleanup(self.curl_multi)
			
			self.curl_multi_running_pointer = None
			self.curl_multi = None
		
		return super()._cleanup_after_perform()

	def _perform_multi_read(self):
		if not self.curl_multi:
			raise Exception("Curl perform is not running.")

		if isinstance(self.curl, curl_cffi.Curl):
			return self._perform_multi_curl_cffi()

		if isinstance(self.curl, pycurl.Curl):
			return self._perform_multi_pycurl()

		raise TypeError("Cannot perform on invalid Curl object.")
	
	def _dequeue_chunks(self):

		if not hasattr(self, '_body_gen'):
			# Body stream
			self._body_gen = self._read_body()
		
		try:
			next(self._body_gen)
		except StopIteration:
			pass

		if self.error:
			raise self.error

		try:
			chunk = self.chunk_queue.get_nowait()
			return chunk
		except _THREAD_QUEUE_MODULE.Empty:
			# No data available anymore, break
			raise QueueContinueRead()

	def _read_headers(self):
		while not self.initialized.is_set():
			self._perform_multi_read()
			yield
			
	def _read_body(self):
		while not self.perform_finished.is_set():
			self._perform_multi_read()
			yield

	def _wait_for_headers(self):
		try:
			for _ in self._read_headers():
				pass
		except StopIteration:
			pass
	
	def _wait_for_body(self):
		try:
			for _ in self._read_body():
				pass
		except StopIteration:
			pass
	
	def _perform(self):
		if self.debug:
			print("[DEBUG] Using Curl Multi Stream Handler.")
			
		# Initialize curl multi
		if isinstance(self.curl, curl_cffi.Curl):

			self.curl.setopt(CurlOpt.WRITEFUNCTION, self._write_callback)
			self.curl._ensure_cacert()

			# Init Multi
			self.curl_multi = lib.curl_multi_init()

			lib.curl_multi_add_handle(self.curl_multi, self.curl._curl)
			# running flag
			self.curl_multi_running_pointer = ffi.new("int *", 0)

		elif isinstance(self.curl, pycurl.Curl):
			self.curl.setopt(pycurl.WRITEFUNCTION, self._write_callback)

			# Init Multi
			self.curl_multi = pycurl.CurlMulti()
			self.curl_multi.add_handle(self.curl)
		else:
			raise TypeError("Cannot perform on invalid Curl object.")



from .base import (
	CurlStreamHandlerBase, 
	QueueContinueRead, 
	QueueBreakRead,
)

from ._thread_env import (
	_THREAD_ENV, _THREAD_CLASS, _THREAD_SLEEP, _THREAD_EVENT, _THREAD_QUEUE_MODULE
)


class CurlStreamHandlerThreads(CurlStreamHandlerBase):
	'''
		Curl Stream Handler (c) 2025 by Elis K.

		Uses threads (either native, or gevent/eventlet) to spawn curl perform in a non-blocking way.
	'''

	def _wait_for_headers(self):
		if self.debug:
			print("[DEBUG] Waiting for headers")
		_done = self.initialized.wait(timeout=self.event_timeout)

		if not _done:
			raise self.read_timeout_error

	def _wait_for_body(self):
		if self.debug:
			print("[DEBUG] Waiting for body")
		self.perform_finished.wait()

	def _perform(self):
		if self.debug:
			print("[DEBUG] Using Curl Threads Stream Handler.")

		from concurrent.futures import ThreadPoolExecutor

		basic_perform = super()._perform

		if _THREAD_ENV == "gevent":
			_THREAD_CLASS.get_hub().threadpool.spawn(basic_perform).get()

		elif _THREAD_ENV == "eventlet":
			_THREAD_CLASS.tpool.execute(basic_perform)
		
		else:
			self.executor = ThreadPoolExecutor()
			self._future = self.executor.submit(basic_perform)

	def close(self):
		if self.closed:
			return
		
		if self._future:
			self._future.result() 

		return super().close()



from concurrent.futures import ThreadPoolExecutor
from io import BytesIO
import math
import os
import re
import threading
import time
import traceback
from typing import TypedDict
import typing
import warnings

from urllib3.util import parse_url
from urllib3.response import HTTPResponse
from urllib3._collections import HTTPHeaderDict

import requests

from requests.utils import (
	DEFAULT_CA_BUNDLE_PATH,
	extract_zipped_paths,
	get_auth_from_url,
	get_encoding_from_headers,
	prepend_scheme_if_needed,
	select_proxy,
	urldefragauth,
)

from requests.compat import urlparse
from requests.structures import CaseInsensitiveDict
from requests.cookies import extract_cookies_to_jar
from requests.adapters import BaseAdapter
from requests.models import Response
from requests.exceptions import (
	RequestException,
	ConnectionError,
	ConnectTimeout,
	InvalidHeader,
	InvalidProxyURL,
	InvalidSchema,
	InvalidURL,
	ProxyError,
	ReadTimeout,
	RetryError,
	SSLError,
	HTTPError,
	TooManyRedirects,
	ContentDecodingError,
	ChunkedEncodingError
)
import pycurl
import curl_cffi.curl
from curl_cffi.curl import CurlInfo as CurlInfoOpt, CurlOpt, CurlError 
from curl_cffi.const import CurlECode, CurlHttpVersion

from .stream.handler import CurlStreamHandler, CurlStreamHandlerBase
from .stream.response import CurlStreamResponse

class CurlInfo(TypedDict):
	local_ip: str
	local_port: int
	primary_ip: str
	primary_port: int

	speed_download: int
	speed_upload: int

	request_size: int
	request_body_size: int
	response_body_size: int
	response_header_size: int

	ssl_verify_result: int
	proxy_ssl_verify_result: int

	total_time: int
	starttransfer_time: int
	connect_time: int
	appconnect_time: int
	pretransfer_time: int
	namelookup_time: int
	has_used_proxy: int


def get_curl_info(response: requests.Response) -> CurlInfo:
	return response.curl_info

class Response(requests.Response):
	curl_info: CurlInfo
	wait_for_body: typing.Callable[[], None]

class BaseCurlAdapter(BaseAdapter):

	def __init__(self, 
		curl_class: typing.Union[curl_cffi.Curl, pycurl.Curl],
		debug=False, 
		use_curl_content_decoding=False,
		use_thread_local_curl=True,
		stream_handler: CurlStreamHandlerBase=None
	):
		self.curl_class: typing.Union[curl_cffi.Curl, pycurl.Curl] = curl_class
		self.debug = debug
		self.use_curl_content_decoding = use_curl_content_decoding
		
		self.use_thread_local_curl = use_thread_local_curl

		self.stream_handler = (stream_handler or CurlStreamHandler)

		if self.use_thread_local_curl:
			self._local = threading.local()
			try:
				from .stream.handler._thread_env import _THREAD_ENV
				if _THREAD_ENV == "gevent":
					from gevent.local import local as gevent_local
					self._local = gevent_local()
			except Exception:
				pass
			self._local.curl = self.curl_class()
		else:
			self._curl = self.curl_class()

		if self.debug:
			self.enable_debug()

	@property
	def curl(self) -> typing.Union[curl_cffi.Curl, pycurl.Curl]:
		if self.use_thread_local_curl:
			if not getattr(self._local, "curl", None):
				self._local.curl = self.curl_class()
			return self._local.curl
		return self._curl
	
	def reset_curl(self, curl: typing.Union[curl_cffi.Curl, pycurl.Curl, None] = None):
		'''
			Close current handle and open a new curl handle
		'''
		if curl is None:
			curl = self.curl
		curl.close()

		if self.use_thread_local_curl:
			self._local.curl = self.curl_class()
			return self._local.curl
		else:
			self._curl = self.curl_class()
			return self._curl

	def enable_debug(self):
		if self.debug:
			self.curl.setopt(CurlOpt.VERBOSE, 1)
		
	def cert_verify(self, curl, url: str, verify: bool, cert):
		"""
		Verify an SSL certificate for HTTPS requests.
		"""

		if url.lower().startswith("https"):
			if verify is True:
				curl.setopt(CurlOpt.SSL_VERIFYPEER, 1)
				curl.setopt(CurlOpt.SSL_VERIFYHOST, 1)
			elif isinstance(verify, str):
				if not os.path.exists(verify):
					raise OSError(
						f"Could not find a suitable TLS CA certificate bundle at: {verify}"
					)
				curl.setopt(CurlOpt.CAINFO, verify)
			else:
				curl.setopt(CurlOpt.SSL_VERIFYPEER, 0)
				curl.setopt(CurlOpt.SSL_VERIFYHOST, 0)

			if cert:
				if isinstance(cert, (list, tuple)) and len(cert) == 2:
					cert_file, key_file = cert
					if not os.path.exists(cert_file):
						raise OSError(
							f"Could not find the TLS certificate file at: {cert_file}"
						)
					if not os.path.exists(key_file):
						raise OSError(f"Could not find the TLS key file at: {key_file}")
					curl.setopt(CurlOpt.SSLCERT, cert_file)
					curl.setopt(CurlOpt.SSLKEY, key_file)
				elif isinstance(cert, str):
					if not os.path.exists(cert):
						raise OSError(
							f"Could not find the TLS certificate file at: {cert}"
						)
					curl.setopt(CurlOpt.SSLCERT, cert)
				else:
					raise ValueError("Invalid SSL certificate format.")
		else:
			curl.setopt(CurlOpt.SSL_VERIFYPEER, 0)
			curl.setopt(CurlOpt.SSL_VERIFYHOST, 0)

	CODE2ERROR = {
			0: RequestException,
			CurlECode.UNSUPPORTED_PROTOCOL: InvalidSchema,
			CurlECode.URL_MALFORMAT: InvalidURL,
			CurlECode.COULDNT_RESOLVE_PROXY: ProxyError,
			CurlECode.COULDNT_RESOLVE_HOST: ConnectionError, #DNSError
			CurlECode.COULDNT_CONNECT: ConnectionError,
			CurlECode.WEIRD_SERVER_REPLY: ConnectionError,
			CurlECode.REMOTE_ACCESS_DENIED: ConnectionError,
			CurlECode.HTTP2: HTTPError,
			CurlECode.HTTP_RETURNED_ERROR: HTTPError,
			CurlECode.WRITE_ERROR: RequestException,
			CurlECode.READ_ERROR: RequestException,
			CurlECode.OUT_OF_MEMORY: RequestException,
			CurlECode.OPERATION_TIMEDOUT: ConnectTimeout, #Timeout,
			CurlECode.SSL_CONNECT_ERROR: SSLError,
			CurlECode.INTERFACE_FAILED: RequestException, #InterfaceError,
			CurlECode.TOO_MANY_REDIRECTS: TooManyRedirects,
			CurlECode.UNKNOWN_OPTION: RequestException,
			CurlECode.SETOPT_OPTION_SYNTAX: RequestException,
			CurlECode.GOT_NOTHING: ConnectionError,
			CurlECode.SSL_ENGINE_NOTFOUND: SSLError,
			CurlECode.SSL_ENGINE_SETFAILED: SSLError,
			CurlECode.SEND_ERROR: ConnectionError,
			CurlECode.RECV_ERROR: ConnectionError,
			CurlECode.SSL_CERTPROBLEM: SSLError,
			CurlECode.SSL_CIPHER: SSLError,
			CurlECode.PEER_FAILED_VERIFICATION: SSLError, #CertificateVerifyError,
			CurlECode.BAD_CONTENT_ENCODING: ContentDecodingError,
			CurlECode.SSL_ENGINE_INITFAILED: SSLError,
			CurlECode.SSL_CACERT_BADFILE: SSLError,
			CurlECode.SSL_CRL_BADFILE: SSLError,
			CurlECode.SSL_ISSUER_ERROR: SSLError,
			CurlECode.SSL_PINNEDPUBKEYNOTMATCH: SSLError,
			CurlECode.SSL_INVALIDCERTSTATUS: SSLError,
			CurlECode.HTTP2_STREAM: HTTPError,
			CurlECode.HTTP3: HTTPError,
			CurlECode.QUIC_CONNECT_ERROR: ConnectionError,
			CurlECode.PROXY: ProxyError,
			CurlECode.SSL_CLIENTCERT: SSLError,
			CurlECode.ECH_REQUIRED: SSLError,
			CurlECode.PARTIAL_FILE: ChunkedEncodingError,
	}

	def curl_error_map(self, error: typing.Union[CurlError, pycurl.error], has_proxy=None):
		
		err_code = 0
		if hasattr(error, 'code'):
			#curl_cffi.CurlError
			err_code = error.code

		elif len(error.args) > 0 and isinstance(error.args[0], int):
			#pycurl.error
			err_code = error.args[0]

		err_message = str(error)

		if err_code == CurlECode.RECV_ERROR and "CONNECT" in err_message and has_proxy:
			return ProxyError

		# Read timeout
		if err_code == 28 and ("Read timeout." in err_message or "Operation too slow." in err_message):
			return requests.exceptions.ReadTimeout
		
		return self.CODE2ERROR.get(err_code, RequestException)

	def get_curl_info(self, curl: typing.Union[curl_cffi.Curl, pycurl.Curl], option_code: int):
		try:
			return curl.getinfo(option_code)
		except:
			if self.debug:
				traceback.print_exc()
			return None
	def parse_info(self, curl: typing.Union[curl_cffi.Curl, pycurl.Curl], headers_only=False):

		additional_info = {
			# IP/Ports
			"local_ip": self.get_curl_info(curl, CurlInfoOpt.LOCAL_IP), 
			"local_port": self.get_curl_info(curl, CurlInfoOpt.LOCAL_PORT), 
			"primary_ip": self.get_curl_info(curl, CurlInfoOpt.PRIMARY_IP), 
			"primary_port": self.get_curl_info(curl, CurlInfoOpt.PRIMARY_PORT), 
			
			# Sizes
			"request_size": self.get_curl_info(curl, CurlInfoOpt.REQUEST_SIZE), # This is always 0 in Curl_Cffi
			"request_body_size": self.get_curl_info(curl, CurlInfoOpt.SIZE_UPLOAD_T), 
			"response_header_size": self.get_curl_info(curl, CurlInfoOpt.HEADER_SIZE),

			# SSL
			"ssl_verify_result": self.get_curl_info(curl, CurlInfoOpt.SSL_VERIFYRESULT),
			"proxy_ssl_verify_result": self.get_curl_info(curl, CurlInfoOpt.PROXY_SSL_VERIFYRESULT),

			# Times
			"starttransfer_time": self.get_curl_info(curl, CurlInfoOpt.STARTTRANSFER_TIME_T),
			"connect_time": self.get_curl_info(curl, CurlInfoOpt.CONNECT_TIME_T),
			"appconnect_time": self.get_curl_info(curl, CurlInfoOpt.APPCONNECT_TIME_T),
			"pretransfer_time": self.get_curl_info(curl, CurlInfoOpt.PRETRANSFER_TIME_T),
			"namelookup_time": self.get_curl_info(curl, CurlInfoOpt.NAMELOOKUP_TIME_T),

			# Other
			"has_used_proxy": self.get_curl_info(curl, CurlInfoOpt.USED_PROXY)
		}

		if not headers_only:
			# Available after the body has been parsed

			additional_info.update({
				"speed_download": self.get_curl_info(curl, CurlInfoOpt.SPEED_DOWNLOAD_T), 
				"speed_upload": self.get_curl_info(curl, CurlInfoOpt.SPEED_UPLOAD_T), 
				"response_body_size": self.get_curl_info(curl, CurlInfoOpt.SIZE_DOWNLOAD_T), 
				"total_time": self.get_curl_info(curl, CurlInfoOpt.TOTAL_TIME_T), 
			})
		
		return additional_info

	def parse_headers(self, curl: typing.Union[curl_cffi.Curl, pycurl.Curl], header_buffer: BytesIO):
		
		def parse_status_line(status_line: str) -> typing.List[str]:

			match = re.match(r"^HTTP/(\d|\d\.\d)\s+([0-9]{3})(?:\s+(.*))?$", status_line)

			if not match:
				return CurlHttpVersion.V1_0, 0, ""

			map_http_versions = {
				"2": CurlHttpVersion.V2_0,
				"2.0": CurlHttpVersion.V2_0,
				"1.1": CurlHttpVersion.V1_1,
				"1.0": CurlHttpVersion.V1_0,
				"1": CurlHttpVersion.V1_0
			}

			_http_version, _status_code, _reason = match.groups()
			
			http_version = (
				map_http_versions[_http_version] if _http_version in map_http_versions else CurlHttpVersion.V1_0
			)

			status_code = int(_status_code)

			reason = _reason or ""
				
			return http_version, status_code, reason

		http_version = CurlHttpVersion.V1_0
		status_code = 0
		reason: str = ""

		header_lines = header_buffer.getvalue().splitlines()
		
		header_list: list[bytes] = []
		for header_line in header_lines:
			if not header_line.strip():
				continue
			if header_line.startswith(b"HTTP/"):

				# read header from last response
				http_version, status_code, reason = parse_status_line(header_line.decode())

				# empty header list for new redirected response
				header_list = []
				continue
			if header_line.startswith(b" ") or header_line.startswith(b"\t"):
				header_list[-1] += header_line
				continue
			header_list.append(header_line)
		
		header_dict = HTTPHeaderDict()

		for header_item in header_list:			
			header_key, header_value = header_item.decode("ascii").split(":", maxsplit=1)
			header_dict.add(
				header_key.strip().title(), header_value.strip()
			)

		return {
			"version": http_version, 
			"status": status_code, 
			"reason": reason,
			"headers": header_dict,
			"header_list": header_list,
		}

	def build_response(self, 
		curl: typing.Union[curl_cffi.Curl, pycurl.Curl], 
		res:CurlStreamResponse, 
		parsed_headers: dict, 
		req: requests.PreparedRequest, 
		wait_for_body: callable,
		curl_info_dict: dict
	) -> Response:
		
		response = Response()

		# Fallback to None if there's no status_code, for whatever reason.
		response.status_code = parsed_headers["status"]

		# Make headers case-insensitive.
		response.headers = CaseInsensitiveDict(parsed_headers["headers"])

		# Set encoding.
		response.encoding = get_encoding_from_headers(response.headers)
		response.raw = res
		
		response.reason = parsed_headers["reason"]

		response.wait_for_body = wait_for_body

		response.curl_info = curl_info_dict

		if isinstance(req.url, bytes):
			response.url = req.url.decode("utf-8")
		else:
			response.url = req.url
		
	
		# Add new cookies from the server.
		extract_cookies_to_jar(response.cookies, req, res)

		# Give the Response some context.
		response.request = req
		response.connection = self

		return response

	def request_url(self, request: requests.PreparedRequest, proxies):
		"""
		Obtain the URL to use when making the final request.
		"""
		'''
		proxy = select_proxy(request.url, proxies)
		scheme = urlparse(request.url).scheme

		is_proxied_http_request = proxy and scheme != "https"
		using_socks_proxy = False
		if proxy:
			proxy_scheme = urlparse(proxy).scheme.lower()
			using_socks_proxy = proxy_scheme.startswith("socks")

		url = request.path_url
		if url.startswith("//"):  # Don't confuse Curl
			url = f"/{url.lstrip('/')}"

		if is_proxied_http_request and not using_socks_proxy:
			url = 

		'''

		return urldefragauth(request.url)

	def _get_request_adapter_options(self, request: requests.PreparedRequest) -> dict:
		"""
			Read per-request adapter flags from internal request headers.
		"""
		options = {}

		header_value = request.headers.pop("X-Curl-Adapter-Disable-Tunnel-Reuse", None)
		if header_value is not None:
			options["disable_tunnel_reuse"] = str(header_value).strip().lower() in (
				"1", "true", "yes", "on"
			)

		return options
	
	def set_curl_options(self, 
			curl: typing.Union[curl_cffi.Curl, pycurl.Curl], 
			request: requests.PreparedRequest,
			url:str, 
			timeout, 
			proxies,
			request_adapter_options=None
		):
		request_adapter_options = request_adapter_options or {}
		disable_tunnel_reuse = bool(request_adapter_options.get("disable_tunnel_reuse", False))
		
		if self.debug:
			print("[DEBUG] Sending: ", url, request.headers, timeout, proxies)
		# url
		curl.setopt(CurlOpt.URL, url.encode())

		# method
		method = request.method.upper()
		if method == "POST":
			curl.setopt(CurlOpt.POST, 1)
		elif method != "GET":
			curl.setopt(CurlOpt.CUSTOMREQUEST, method.encode())
		if method == "HEAD":
			curl.setopt(CurlOpt.NOBODY, 1)

		# timeout 
		if timeout is None:
			timeout = 0  # indefinitely
		
		if isinstance(timeout, tuple):
			connect_timeout, read_timeout = timeout

			curl.setopt(CurlOpt.CONNECTTIMEOUT_MS, int(connect_timeout * 1000))
	
			#curl.setopt(CurlOpt.TIMEOUT_MS, int(all_timeout * 1000))
		
			# trick from: https://github.com/lexiforest/curl_cffi/issues/156
			curl.setopt(CurlOpt.LOW_SPEED_LIMIT, 1)
			curl.setopt(CurlOpt.LOW_SPEED_TIME, math.ceil(read_timeout))
	
		elif isinstance(timeout, (int, float)):

				#curl.setopt(CurlOpt.TIMEOUT_MS, int(timeout * 1000))
		
				curl.setopt(CurlOpt.CONNECTTIMEOUT_MS, int(timeout * 1000))
				curl.setopt(CurlOpt.LOW_SPEED_LIMIT, 1)
				curl.setopt(CurlOpt.LOW_SPEED_TIME, math.ceil(timeout))

		# body
		body = b"" if not request.body else (
			request.body.encode() if isinstance(request.body, str) else request.body
		)

		if body or method in ("POST", "PUT", "PATCH"):
			curl.setopt(CurlOpt.POSTFIELDS, body)
			# necessary if body contains '\0'
			curl.setopt(CurlOpt.POSTFIELDSIZE, len(body))
			if method == "GET":
				curl.setopt(CurlOpt.CUSTOMREQUEST, method)
	
		# headers
		host_header = request.headers.get("host")
		if host_header is not None:
			# remove Host header if it's unnecessary, otherwise curl may get confused.
			# Host header will be automatically added by curl if it's not present.
			# https://github.com/lexiforest/curl_cffi/issues/119
			parsed_url = urlparse(url)
			if host_header == parsed_url.netloc or host_header == parsed_url.hostname:
				request.headers.pop("Host", None)

		request.headers.pop("Expect", None) # Never send `Expect` header. ?
		
		header_lines = []
		for k, v in request.headers.items():
			# Make Curl Headers Array
			# Make curl always include empty headers.
			# See: https://stackoverflow.com/a/32911474/1061155

			if v is None:
				header_lines.append(f"{k}:".encode())  # Explictly disable this header
			elif v == "":
				header_lines.append(f"{k};".encode())  # Add an empty valued header
			else:
				header_lines.append(f"{k}: {v}".encode())
	
		curl.setopt(CurlOpt.HTTPHEADER, header_lines)
		
		# cookies
		#already handled
	
		# files
		#already handled

		# auth
		#already handled, it's just a header...
		
		# allow_redirects
		curl.setopt(CurlOpt.FOLLOWLOCATION, 0)  # Don't allow. Requests library handles them by itself.
	
		# proxies
		proxy = select_proxy(request.url, proxies)
		if proxy:
			proxy: str = prepend_scheme_if_needed(proxy, "http")
			proxy_url = parse_url(proxy)
			if not proxy_url.host:
				raise InvalidProxyURL(
					"Please check proxy URL. It is malformed "
					"and could be missing the host."
				)


			curl.setopt(CurlOpt.PROXY, proxy)

			# Authentication?
			username, password = get_auth_from_url(proxy)
			if username:
				curl.setopt(CurlOpt.PROXYUSERNAME, username.encode())
				curl.setopt(CurlOpt.PROXYPASSWORD, password.encode())

			if proxy.lower().startswith("socks"):
				pass
			else:
				if proxy_url.scheme == "https":
					# For https site with http tunnel proxy, tell curl to enable tunneling
					curl.setopt(CurlOpt.HTTPPROXYTUNNEL, 1)

			if disable_tunnel_reuse:
				curl.setopt(CurlOpt.FRESH_CONNECT, 1)
				curl.setopt(CurlOpt.FORBID_REUSE, 1)
	
		
		# content decoding
		if self.use_curl_content_decoding:
			curl.setopt(CurlOpt.HTTP_CONTENT_DECODING, 1)
			curl.setopt(CurlOpt.ACCEPT_ENCODING, "gzip, deflate, br, zstd")
		else:
			curl.setopt(CurlOpt.HTTP_CONTENT_DECODING, 0)
			curl.setopt(CurlOpt.HTTP_TRANSFER_DECODING, 1)

		# do not check max_recv_speed
		curl.setopt(CurlOpt.MAX_RECV_SPEED_LARGE, 0)

	def send(
		self, request: requests.PreparedRequest, stream=False, timeout=None, verify=True, cert=None, proxies=None
	):
		curl = self.reset_curl()
		
		self.cert_verify(curl, request.url, verify, cert)

		url = self.request_url(request, proxies)
		request_adapter_options = self._get_request_adapter_options(request)

		self.set_curl_options(
			curl,
			request=request,
			url=url,
			timeout=timeout,
			proxies=proxies,
			request_adapter_options=request_adapter_options
		)
		a = time.time()
		try:
			# Save headers when received
			header_buffer = BytesIO()
			curl.setopt(CurlOpt.HEADERDATA, header_buffer)

			# Callbacks for retrieving & saving curl info object
			curl_info_dict: CurlInfo = {}

			# Perform curl request with threading, and return body in a 'read' like class type (by simply using Curl.WRITEFUNCTION callback)
			start_curl_stream = (
				self.stream_handler(
					curl_instance=curl,
					callback_after_perform=lambda curl: curl_info_dict.update(self.parse_info(curl)),
					timeout=timeout,
					debug=self.debug
				)
			).start()
			
			# Headers are already available
			curl_info_dict.update(self.parse_info(curl, headers_only=True))

			if self.debug:
				print("[DEBUG] Curl Start Elapsed Time: ", time.time() - a)

			# Headers are available after start, parse them
			parsed_headers = self.parse_headers(curl, header_buffer)

			curl_stream_res = CurlStreamResponse(
				url=url,
				method=request.method.upper(),
				request=request,
				curl_stream_handler=start_curl_stream,
				use_curl_content_decoding=self.use_curl_content_decoding,
				**parsed_headers
			)

			return self.build_response(curl, curl_stream_res, parsed_headers, request, wait_for_body=start_curl_stream._wait_for_body, curl_info_dict=curl_info_dict)
		except OSError as e:
			raise ConnectionError(e, request=request)
		
		except (CurlError, pycurl.error) as e:
			error_to_throw = self.curl_error_map(e, has_proxy=proxies)
			raise error_to_throw(e, request=request)

		except Exception as e:
			if self.debug:
				traceback.print_exc()
			raise e
		finally:
			if self.debug:
				print("[DEBUG] Curl Send Elapsed Time: ", time.time() - a)
			pass
		
	def close(self) -> None:
		"""Close the session."""
		self._closed = True
		self.curl.close()

	def __enter__(self):
		return self

	def __exit__(self, *args):
		self.close()




import typing
import warnings
from typing import TYPE_CHECKING, Any

from curl_cffi._wrapper import ffi, lib
from curl_cffi.const import CurlMOpt
from curl_cffi.curl import Curl
from curl_cffi.utils import CurlCffiWarning

import gevent
from gevent.event import AsyncResult
from gevent.lock import Semaphore

if TYPE_CHECKING:
	from gevent._types import _IoWatcher # type: ignore

CURL_POLL_NONE = 0
CURL_POLL_IN = 1
CURL_POLL_OUT = 2
CURL_POLL_INOUT = 3
CURL_POLL_REMOVE = 4

CURL_SOCKET_TIMEOUT = -1
CURL_SOCKET_BAD = -1

CURL_CSELECT_IN = 0x01
CURL_CSELECT_OUT = 0x02
CURL_CSELECT_ERR = 0x04

CURLMSG_DONE = 1

GEVENT_READ = 1
GEVENT_WRITE = 2

_TIMER_CALLBACK_RETURNS_INT = ffi.typeof(lib.timer_function).result.cname != "void"
_SOCKET_CALLBACK_RETURNS_INT = ffi.typeof(lib.socket_function).result.cname != "void"


def _timer_callback_return():
	return 0 if _TIMER_CALLBACK_RETURNS_INT else None


def _socket_callback_return():
	return 0 if _SOCKET_CALLBACK_RETURNS_INT else None


@ffi.def_extern()
def timer_function(curlm, timeout_ms: int, clientp: "GeventCurlCffi"):
	try:
		gevent_curl: "GeventCurlCffi" = ffi.from_handle(clientp)

		if gevent_curl._timer:
			gevent_curl._timer.kill(block=False)
			gevent_curl._timer = None

		if not gevent_curl._curl_multi:
			# Don't schedule timers if multi handle is gone.
			return _timer_callback_return()

		if timeout_ms == -1:
			# A timeout_ms value of -1 means you should delete the timer, we already did.
			return _timer_callback_return()

		if timeout_ms == 0:
			# immediate timeout; invoke directly
			gevent_curl._timer = gevent.spawn(
				gevent_curl._process_data,
				CURL_SOCKET_TIMEOUT,
				CURL_POLL_NONE
			)
		elif timeout_ms > 0:
			# spawn a greenlet to run after timeout_ms milliseconds
			gevent_curl._timer = gevent.spawn_later(
				timeout_ms / 1000.0,
				gevent_curl._process_data,
				CURL_SOCKET_TIMEOUT,
				CURL_POLL_NONE,
			)
	except Exception:
		# Never raise from CFFI callbacks; Windows can show "Python-CFFI error" dialogs.
		return _timer_callback_return()

	return _timer_callback_return()

@ffi.def_extern()
def socket_function(curlm, sockfd: int, what: int, clientp: "GeventCurlCffi", data: Any):
	try:
		gevent_curl: "GeventCurlCffi" = ffi.from_handle(clientp)
		want_read = bool(what & CURL_POLL_IN)
		want_write = bool(what & CURL_POLL_OUT)

		# compute the new mask for gevent
		new_mask = 0
		if want_read:
			new_mask |= GEVENT_READ
		if want_write:
			new_mask |= GEVENT_WRITE

		if what & CURL_POLL_REMOVE:
			gevent_curl._update_watcher(sockfd, 0)
			return _socket_callback_return()

		gevent_curl._update_watcher(sockfd, new_mask)
	except Exception:
		# Never raise from CFFI callbacks; Windows can show "Python-CFFI error" dialogs.
		return _socket_callback_return()

	return _socket_callback_return()

class GeventCurlCffi:
	'''
		Usage:

		multi_curl = GeventCurlCffi()
		result = multi_curl.add_handle(curl_handle)

		result.wait()
	'''

	def __init__(self):
		"""
		Parameters:
			cacert: CA cert path to use, by default, certs from ``certifi`` are used.
			loop: EventLoop to use.
		"""
		self._curl_multi = lib.curl_multi_init()
		
		self.loop = gevent.get_hub().loop

		self._timer: typing.Optional[gevent.Greenlet] = None
		self._watchers: dict[typing.Any, dict[str, _IoWatcher]] = {}

		self._results: dict[Curl, AsyncResult] = {}
		self._handles: dict[ffi.CData, Curl] = {}
		self._callbacks: dict[Curl, typing.Callable] = {}
		
		self._checker = gevent.spawn(self._force_timeout)

		self._start_closing = False

		self._set_options()

	def add_handle(self, curl: Curl, cleanup_after_perform: typing.Callable[[typing.Optional[Exception]], None]=None):
		"""Add a curl handle to be managed by curl_multi. This is the equivalent of
		`perform` in the async world."""

		if self._start_closing:
			raise RuntimeError("This curl_multi instance is closed.")

		code = lib.curl_multi_add_handle(self._curl_multi, curl._curl)
		
		if code != 0:
			curl_error = curl._get_error(code, "perform")
			self._set_exception(curl, RuntimeError(f"curl_multi_add_handle failed: {curl_error}"))
			
			# return a future that’s already failed
			result = AsyncResult()
			result.set_exception(RuntimeError(curl_error))
			return result
		
		result = AsyncResult()
		self._results[curl] = result
		self._callbacks[curl] = cleanup_after_perform
		self._handles[curl._curl] = curl

		return result

	def cancel_handle(self, curl: Curl):
		"""Cancel is not natively supported in gevent.AsyncResult."""

		# No true cancellation; set an exception or drop reference
		self._set_exception(curl, RuntimeError("Cancelled"))

	def graceful_close(self):
		self._start_closing = True

	def close(self):
		"""Close and cleanup running timers, readers, writers and handles."""
		 # Close and wait for the force timeout checker to complete

		if self._checker and not self._checker.dead:
			self._checker.kill(block=False)

		# Close all pending futures
		for curl in list(self._results.keys()):
			self.cancel_handle(curl)
			
		# Cleanup curl_multi handle
		if self._curl_multi:
			ref_curl_multi = self._curl_multi
			self._curl_multi = None
			lib.curl_multi_cleanup(ref_curl_multi)
			ref_curl_multi = None
			

		# Remove watchers
		for sockfd, entry in list(self._watchers.items()):
			if entry.get("watcher"):
				entry.get("watcher").stop()      # stop monitoring
				entry.get("watcher").close()      # dispose of the watcher
				del self._watchers[sockfd]

		# Kill timer
		if self._timer is not None:
			self._timer.kill(block=False)
			self._timer = None

	def _set_options(self):
		lib.curl_multi_setopt(self._curl_multi, CurlMOpt.TIMERFUNCTION, lib.timer_function)
		lib.curl_multi_setopt(self._curl_multi, CurlMOpt.SOCKETFUNCTION, lib.socket_function)

		self._self_handle = ffi.new_handle(self)
		lib.curl_multi_setopt(self._curl_multi, CurlMOpt.SOCKETDATA, self._self_handle)
		lib.curl_multi_setopt(self._curl_multi, CurlMOpt.TIMERDATA, self._self_handle)

	def _socket_action(self, sockfd: int, ev_bitmask: int) -> int:
		"""Call libcurl socket_action function"""
		running_handle = ffi.new("int *")
		lib.curl_multi_socket_action(self._curl_multi, sockfd, ev_bitmask, running_handle)
		return running_handle[0]
	
	def _process_data(self, sockfd: int, ev_bitmask: int):
		"""Call curl_multi_info_read to read data for given socket."""
		if not self._curl_multi:
			warnings.warn(
				"Curlm already closed! quitting from _process_data",
				CurlCffiWarning,
				stacklevel=2,
			)
			return

		self._socket_action(sockfd, ev_bitmask)

		msg_in_queue = ffi.new("int *")
		while True:
			if not self._curl_multi:
				break
			
			curl_msg = lib.curl_multi_info_read(self._curl_multi, msg_in_queue)
			# print("message in queue", msg_in_queue[0], curl_msg)
			if curl_msg == ffi.NULL:
				break
			if curl_msg.msg == CURLMSG_DONE:
				# print("curl_message", curl_msg.msg, curl_msg.data.result)
				curl = self._handles[curl_msg.easy_handle]
				retcode = curl_msg.data.result
				curl_error = None
				if retcode == 0:
					self._set_result(curl)
				else:
					curl_error = curl._get_error(retcode, "perform")
					self._set_exception(curl, curl_error)
	
	def _force_timeout(self):
		while self._curl_multi:
			gevent.sleep(1)
			# Drive socket action and drain completion/error messages.
			# Without draining info_read, a transfer can finish without resolving its AsyncResult.
			self._process_data(CURL_SOCKET_TIMEOUT, CURL_POLL_NONE)

	def _pop_future(self, curl: Curl):
		lib.curl_multi_remove_handle(self._curl_multi, curl._curl)
		self._handles.pop(curl._curl, None)
		return self._results.pop(curl, None)
	
	def _callback(self, curl: Curl, error: Exception=None):
		if curl in self._callbacks:
			callback = self._callbacks.pop(curl)
			if callable(callback):
				callback(error)
		
	def _set_result(self, curl: Curl):
		result = self._pop_future(curl)
		self._callback(curl)
		if result and not result.ready():
			result.set(None)

		if self._start_closing and not self._results:
			self.close()
		
	def _set_exception(self, curl: Curl, exception):
		result = self._pop_future(curl)
		self._callback(curl, exception)
		
		if result and not result.ready():
			result.set_exception(exception)

		if self._start_closing and not self._results:
			self.close()

	def _update_watcher(self, fd: int, mask: int):
		"""
		Ensure there's exactly one I/O watcher for `fd` with the given mask.
		If mask==0, we stop+remove it. If mask changes, we stop+recreate.
		"""
		entry = self._watchers.get(fd)

		# nothing to do if mask didn’t change
		if entry and entry["mask"] == mask:
			return

		# stop old watcher if any
		if entry:
			entry["watcher"].stop()
			entry["watcher"].close()
			del self._watchers[fd]

		# if new mask is zero, we’re done
		if mask == 0:
			return

		# create a new watcher for read/write as needed
		w = self.loop.io(fd, mask, ref=True, priority=None)
		# callback only gets fd; we’ll re-derive the libcurl bitmask from mask
		w.start(self._on_watcher_event, fd)

		# stash both watcher _and_ the mask we asked for
		self._watchers[fd] = {"watcher": w, "mask": mask}

	def _on_watcher_event(self, fd: int):
		"""
		A gevent-watcher fired on `fd`.  Look up its mask and call curl.
		"""
		entry = self._watchers.get(fd)
		if not entry:
			return

		mask = entry["mask"]
		ev_bitmask = 0
		if mask & GEVENT_READ:
			ev_bitmask |= CURL_CSELECT_IN
		if mask & GEVENT_WRITE:
			ev_bitmask |= CURL_CSELECT_OUT

		# this is exactly what you were doing before:
		# call socket_action + info_read → _process_data
		self._process_data(fd, ev_bitmask)



import queue
import sys
import threading
import time
import typing


def _detect_environment() -> typing.Tuple:
	## -eventlet-
	if "eventlet" in sys.modules:
		try:
			import eventlet
			import eventlet.queue
			from eventlet.patcher import is_monkey_patched as is_eventlet
			import socket
			import eventlet.event

			if is_eventlet(socket):
				return ("eventlet", eventlet, eventlet.sleep, eventlet.event.Event, eventlet.queue)

		except ImportError:
			pass

	# -gevent-
	if "gevent" in sys.modules:
		try:
			import gevent
			import gevent.queue
			from gevent import socket as _gsocket
			import socket
			import gevent.event

			if socket.socket is _gsocket.socket:
				return ("gevent", gevent, gevent.sleep, gevent.event.Event, gevent.queue)
		except ImportError:
			pass

	return ("default", threading, time.sleep, threading.Event, queue)

_THREAD_ENV, _THREAD_CLASS, _THREAD_SLEEP, _THREAD_EVENT, _THREAD_QUEUE_MODULE = _detect_environment()


import gevent.event
import gevent.queue
from gevent.event import AsyncResult
from gevent.lock import Semaphore

from curl_cffi.curl import CurlOpt

from curl_adapter.stream.sockets.curl_cffi_socket import GeventCurlCffi
from curl_adapter.stream.sockets.pycurl_socket import GeventPyCurl

from .base import (
	CurlStreamHandlerBase, 
	QueueContinueRead, 
	QueueBreakRead,
)

class CurlStreamHandlerGevent(CurlStreamHandlerBase):
	'''
		Curl Stream Handler (c) 2025 by Elis K.

		Gevent only. Uses low-level curl socket handlers & multi interface.
	'''
	
	gevent_curl_cffi = GeventCurlCffi()
	gevent_pycurl = GeventPyCurl()

	_requests = 0 # track the number of requests handled
	_rotate_every = 1000 # create a new multi handle every 1000 requests
	_lock = Semaphore()

	def __init__(self, curl_instance, callback_after_perform=None, timeout=None, debug=False):
		
		super().__init__(curl_instance, callback_after_perform, timeout, debug)
		
		# Events
		self.quit_event = gevent.event.Event()  # Signal to stop streaming
		self.initialized = gevent.event.Event() # Event to set when we receive the first bytes of body, that's how we know that the headers are ready
		self.perform_finished = gevent.event.Event() # Body has finished reading

		self._future = None

		self.chunk_queue = gevent.queue.Queue()

	def _wait_for_headers(self):
		_done = self.initialized.wait(timeout=self.event_timeout)

		if not _done:
			raise self.read_timeout_error
		
		if self.debug:
			print("[DEBUG] Headers received.")
	
	def _wait_for_body(self):
		self.perform_finished.wait()

	def _dequeue_chunks(self):
		
		try:
			chunk = self.chunk_queue.get(timeout=1)
			return chunk
		except gevent.queue.Empty:
			raise QueueBreakRead()

	def _perform(self):
		if self.debug:
			print("[DEBUG] Using Gevent Stream Handler.")
		

		if self.curl_type == "curl_cffi":
			self.curl.setopt(CurlOpt.WRITEFUNCTION, self._write_callback)
			self.curl._ensure_cacert()

			self._future: AsyncResult = self.gevent_curl_cffi.add_handle(
				self.curl,
				cleanup_after_perform=self._cleanup_after_perform
			)

		elif self.curl_type == "pycurl":
			self.curl.setopt(CurlOpt.WRITEFUNCTION, self._write_callback)

			self._future = self.gevent_pycurl.add_handle(
				self.curl,
				cleanup_after_perform=self._cleanup_after_perform
			)
	
	@classmethod
	def check_rotate(cls):
		'''
			Re-initialize the `curl_multi` instances every once in a while
		'''
		with cls._lock:
			cls._requests += 1
			if cls._requests >= cls._rotate_every:
				old_gevent_curl_cffi = cls.gevent_curl_cffi
				old_gevent_pycurl = cls.gevent_pycurl

				cls.gevent_curl_cffi = GeventCurlCffi()
				cls.gevent_pycurl = GeventPyCurl()

				old_gevent_curl_cffi.graceful_close()
				old_gevent_pycurl.graceful_close()

				cls._requests = 0
				return True
			return False
		
	def close(self):
		if self.closed:
			return
		if self.debug:
			print("[DEBUG] Starting to close...")
		
		if self._future and not self._future.ready():
			if self.curl_type == "curl_cffi":
				self.gevent_curl_cffi.cancel_handle(self.curl)
			elif self.curl_type == "pycurl":
				self.gevent_pycurl.cancel_handle(self.curl)
			self._future.result()

		rotate_multi = self.__class__.check_rotate()
		if rotate_multi and self.debug:
			print("[DEBUG] Re-initalized the curl multi stream handler instance.")

		if self.debug:
			print("[DEBUG] Closing.")

		return super().close()



"""
Curl-Adapter
A curl HTTP adapter for requests library

:copyright: (c) 2025 by Elis K.
:license: MIT.
"""

__all__ = [
    "CurlCffiAdapter",
    "PyCurlAdapter",
    "CurlInfo",
    "get_curl_info"
]


from .base_adapter import CurlInfo, get_curl_info
from .curl_cffi import CurlCffiAdapter
from .pycurl import PyCurlAdapter
from importlib import metadata

__title__ = "curl_adapter"
__description__ = metadata.metadata("curl_adapter")["Summary"]
__version__ = metadata.version("curl_adapter")


import queue
import threading
import typing
import traceback

from curl_cffi import CurlECode, CurlError
import pycurl
import curl_cffi.curl
from curl_cffi.curl import CurlOpt

from ._thread_env import (
	_THREAD_ENV, _THREAD_CLASS, _THREAD_SLEEP, _THREAD_EVENT, _THREAD_QUEUE_MODULE
)

class QueueBreakRead(Exception):
	pass

class QueueContinueRead(Exception):
	pass

class CurlStreamHandlerBase():
	'''
		Curl Stream Handler (c) 2025 by Elis K.
	'''

	def __init__(self, 
		curl_instance: typing.Union[curl_cffi.Curl, pycurl.Curl], 
		callback_after_perform: typing.Callable[[typing.Union[curl_cffi.Curl, pycurl.Curl]], None]=None, 
		timeout: typing.Union[float, int, typing.Tuple[float, float], None]=None, 
		debug: bool=False
	):
		'''
			Initialize the stream handler.
		'''
		self.curl = curl_instance
		
		self.curl_type: typing.Literal["curl_cffi", "pycurl"] = "curl_cffi" if isinstance(self.curl, curl_cffi.Curl) else (
			 "pycurl" if isinstance(self.curl, pycurl.Curl) else None
		)

		self.read_timeout = timeout[1] if isinstance(timeout, tuple) else timeout

		self.event_timeout = self.read_timeout + 5 if self.read_timeout else None
		'''
			Timeout 'wait' events once the read timeout has already passed. To avoid blocking forever.
		'''

		self.read_timeout_error = (
			CurlError("Read timeout.", CurlECode.OPERATION_TIMEDOUT) if self.curl_type == "curl_cffi" 
			else pycurl.error(28, "Read timeout.")
		)
		
		if not self.curl_type:
			raise TypeError("Invalid curl class object.")
		
		# Events
		self.quit_event: threading.Event = _THREAD_EVENT()
		'''
			Signal to stop
		'''
		
		self.initialized: threading.Event = _THREAD_EVENT() 
		'''
			Event to set when we receive the first bytes of body, that's how we know that the headers are ready
		'''

		self.perform_finished: threading.Event = _THREAD_EVENT()
		'''
			Body has finished reading
		'''

		# Body streaming
		self.chunk_queue: queue.Queue = _THREAD_QUEUE_MODULE.Queue()
		'''
			Thread-safe queue for streaming data
		'''

		self._leftover = bytearray()
		'''
			buffer for leftover data when chunk > requested
		'''

		self.callback_after_perform = callback_after_perform
		self.debug = debug

		self.error = None 
		self.closed = False

	def _write_callback(self, chunk):
		'''
			Callback to handle incoming data chunks.
		'''	
		self.initialized.set()
			
		if self.quit_event.is_set():
			return -1  # Signal to stop

		self.chunk_queue.put(chunk)  # Add chunk to the queue
		return len(chunk)

	def _cleanup_after_perform(self, curl_error=None):

		if curl_error:
			self.error = curl_error

		# signal end of stream
		self.chunk_queue.put(None)
		if callable(self.callback_after_perform):
			try:
				self.callback_after_perform(self.curl)
			except Exception:
				if self.debug:
					traceback.print_exc()
		
		self.perform_finished.set()
		self.initialized.set()
	
	def _wait_for_headers(self):
		pass

	def _wait_for_body(self):
		pass

	def _perform(self):
		'''
			The initial basic, blocking way to call perform.
		'''

		try:
			self.curl.setopt(CurlOpt.WRITEFUNCTION, self._write_callback)
			self.curl.perform()
		except Exception as e: #(CurlError, pycurl.error)
			self.error = e
		finally:
			self._cleanup_after_perform()
		
	def start(self):
		self._perform()
		self._wait_for_headers()

		return self

	def _dequeue_chunks(self):
		try:
			chunk = self.chunk_queue.get(timeout=1)
			return chunk
		except _THREAD_QUEUE_MODULE.Empty:
			raise QueueBreakRead()

	def read(self, amt=None):
		"""
			A more 'file-like' read from the queue:

			- If `amt` is None, read all.
			- If `amt` is an integer, read exactly `amt` bytes.
			- Handles leftover data from previous chunk to avoid losing bytes.
		"""
		if self.closed:
			return b""

		if self.error:
			raise self.error

		# If amt is None, read everything:
		if amt is None:
			return self._read_all()

		# If amt is specified (and possibly 0 or > 0)
		return self._read_amt(amt)

	def _read_all(self):
		"""
			Read *all* remaining data from leftover + queue
		"""
		out = bytearray()

		# If there's leftover data, use it first
		out.extend(self._leftover)
		self._leftover.clear()

		# Then read new chunks until we hit None or are closed
		while not self.closed and not self.quit_event.is_set():
			if self.error:
				raise self.error
			
			try:
				chunk = self._dequeue_chunks()
			except QueueBreakRead:
				break
			except QueueContinueRead:
				continue
			
			if chunk is None:
				# End of stream. Close here?
				if self.perform_finished.is_set():
					self.close()
				break

			out.extend(chunk)
		return bytes(out)

	def _read_amt(self, amt):
		"""
			Read exactly `amt` bytes. Returns up to `amt`.
		"""
		out = bytearray()
		needed = amt

		# First, consume leftover if available
		if self._leftover:
			take = min(needed, len(self._leftover))
			out.extend(self._leftover[:take])
			del self._leftover[:take]
			needed -= take

		
		# Read additional chunks from the queue if we still need data
		while needed > 0 and not self.closed and not self.quit_event.is_set():
			if self.error:
				raise self.error

			try:
				chunk = self._dequeue_chunks()
			except QueueBreakRead:
				break
			except QueueContinueRead:
				continue

			if chunk is None:
				# End of stream. close here?
				if self.perform_finished.is_set():
					self.close()
				break

			# If the chunk is bigger than needed, take part of it
			# and store the remainder in _leftover.
			if len(chunk) > needed:
				out.extend(chunk[:needed])
				self._leftover.extend(chunk[needed:])
				needed = 0
			else:
				# Chunk fits entirely
				out.extend(chunk)
				needed -= len(chunk)
		
		return bytes(out)

	def flush(self):
		pass
	
	def close(self):
		'''
			Signal to stop the streaming and wait for the task to complete.
		'''
		if self.closed:
			return

		self.quit_event.set()
		
		if not self.perform_finished.is_set():
			raise Exception("Curl perform is not finished yet, cannot close.")
		
		self.closed = True

	def __del__(self):
		'''
			Destructor to ensure the response is properly closed when garbage-collected.
		'''
		if not self.closed:
			self.close()
	
	def __exit__(self, *args):
		if not self.closed:
			self.close()



import pycurl
from .base_adapter import BaseCurlAdapter
from .stream.handler.base import CurlStreamHandlerBase

class PyCurlAdapter(BaseCurlAdapter):

	def __init__(self, 
			debug=False, 
			use_curl_content_decoding=False, # pyCurl automatic decoding is disabled by default. Because pycurl doesnt support modern decoding algorithms...
			use_thread_local_curl=True,
			stream_handler: CurlStreamHandlerBase=None
        ):

		super().__init__(
			pycurl.Curl, 
			debug,
			use_curl_content_decoding, 
			use_thread_local_curl,
			stream_handler
		)

	def parse_info(self, curl: pycurl.Curl, headers_only=False):
		'''
		 PyCurl currently doesn't support newer methods like TOTAL_TIME_T, SPEED_DOWNLOAD_T, SPEED_UPLOAD_T, SIZE_UPLOAD_T, SIZE_DOWNLOAD_T, etc.
		 So we use the deprecated ones instead.
		'''
		
		additional_info = {
			# IP/Ports
			"local_ip": self.get_curl_info(curl, pycurl.LOCAL_IP), 
			"local_port": self.get_curl_info(curl, pycurl.LOCAL_PORT), 
			"primary_ip": self.get_curl_info(curl, pycurl.PRIMARY_IP), 
			"primary_port": self.get_curl_info(curl, pycurl.PRIMARY_PORT), 
			
			# Sizes
			"request_size": self.get_curl_info(curl, pycurl.REQUEST_SIZE), 
			"request_body_size": self.get_curl_info(curl, pycurl.SIZE_UPLOAD), 
			"response_header_size": self.get_curl_info(curl, pycurl.HEADER_SIZE),

			# SSL
			"ssl_verify_result": self.get_curl_info(curl, pycurl.SSL_VERIFYRESULT),
			"proxy_ssl_verify_result": "unsupported",

			# Times
			"starttransfer_time": self.get_curl_info(curl, pycurl.STARTTRANSFER_TIME),
			"connect_time": self.get_curl_info(curl, pycurl.CONNECT_TIME),
			"appconnect_time": self.get_curl_info(curl, pycurl.APPCONNECT_TIME),
			"pretransfer_time": self.get_curl_info(curl, pycurl.PRETRANSFER_TIME),
			"namelookup_time": self.get_curl_info(curl, pycurl.NAMELOOKUP_TIME),

			# Other
			"has_used_proxy": "unsupported", 
		}

		if not headers_only:
			# Available after the body has been parsed
			additional_info.update({
				"speed_download": self.get_curl_info(curl, pycurl.SPEED_DOWNLOAD), 
				"speed_upload": self.get_curl_info(curl, pycurl.SPEED_UPLOAD), 
				"response_body_size": self.get_curl_info(curl, pycurl.SIZE_DOWNLOAD), 
				"total_time": self.get_curl_info(curl, pycurl.TOTAL_TIME)
			})
			
		filtered_keys = {k:v for k,v in additional_info.items() if v is not None}

		return filtered_keys